import copy
import glob
import os
import platform
import re
import shlex
import warnings
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import timedelta
from enum import Enum
from typing import Dict, List, Optional, Union

from flyteidl.core.literals_pb2 import KeyValuePair
from flyteidl.core.tasks_pb2 import Container, ContainerPort, ExtendedResources, K8sObjectMetadata, K8sPod
from flyteidl.core.tasks_pb2 import Resources as ResourcesIDL
from flytekit import FlyteContextManager, ImageSpec, Resources
from flytekit.core.artifact import ArtifactQuery
from flytekit.core.pod_template import PodTemplate
from flytekit.extras.accelerators import BaseAccelerator
from flytekit.models.security import Secret
from mashumaro.codecs.json import JSONEncoder

from union.app._common import patch_get_flytekit_for_pypi, patch_get_unionai_for_pypi
from union.internal.app.app_definition_pb2 import App as AppIDL
from union.internal.app.app_definition_pb2 import (
    AutoscalingConfig,
    Identifier,
    IngressConfig,
    Meta,
    Replicas,
    SecurityContext,
    Spec,
)
from union.internal.app.app_definition_pb2 import Concurrency as ConcurrencyIDL
from union.internal.app.app_definition_pb2 import RequestRate as RequestRateIDL
from union.internal.app.app_definition_pb2 import ScalingMetric as ScalingMetricIDL
from union.ucimage._image_builder import get_image_name, is_union_image


@dataclass
class AppRegistry:
    """Keeps track of all user defined applications"""

    apps: Dict[str, "App"] = field(default_factory=dict)

    def add_app(self, app: "App"):
        execution_state = FlyteContextManager.current_context().execution_state
        if execution_state.mode is None and app.name in self.apps:
            # This is running in a non execute context, so we check if the app name is already used:
            msg = f"App with name: {app.name} was already used. The latest definition will be used"
            warnings.warn(msg, UserWarning, stacklevel=2)

        self.apps[app.name] = app


APP_REGISTRY = AppRegistry()
# TODO: Add more protocols here when we deploy to another cloud providers
SUPPORTED_FS_PROTOCOLS = ["s3://", "gs://", "union://", "ufs://", "unionmeta://", "ums://"]
INVALID_APP_PORTS = [8012, 8022, 8112, 9090, 9091]
_PACKAGE_NAME_RE = re.compile(r"^[\w-]+")
_UNION_RUNTIME_NAME = "union-runtime"


def _is_union_runtime(package: str) -> bool:
    """Return True if `package` is union-runtime."""
    m = _PACKAGE_NAME_RE.match(package)
    if not m:
        return False
    name = m.group()
    return name == _UNION_RUNTIME_NAME


def _convert_union_runtime_to_serverless(packages: Optional[list[str]]) -> Optional[list[str]]:
    """Convert packages using union-runtime to union-runtime[serverless]"""
    if packages is None:
        return None

    union_runtime_length = len(_UNION_RUNTIME_NAME)
    new_packages = []
    for package in packages:
        if _is_union_runtime(package):
            version_spec = package[union_runtime_length:]
            new_packages.append(f"union-runtime[serverless]{version_spec}")
        else:
            new_packages.append(package)
    return new_packages


@dataclass
class URLQuery:
    name: str
    public: bool


ENV_NAME_RE = re.compile("^[_a-zA-Z][_a-zA-Z0-9]*$")
APP_NAME_RE = re.compile(r"[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*")


def _has_file_extension(file_path) -> bool:
    _, ext = os.path.splitext(file_path)
    return bool(ext)


@dataclass
class Input:
    """
    Input for application.

    :param name: Name of input.
    :param value: Value for input.
    :param env_var: Environment name to set the value in the serving environment.
    :param download: When True, the input will be automatically downloaded. This
        only works if the value refers to an item in a object store. i.e. `s3://...`
    :param mount: If `value` is a directory, then the directory will be available
        at `mount`. If `value` is a file, then the file will be downloaded into the
        `mount` directory.
    :param ignore_patterns: If `value` is a directory, then this is a list of glob
        patterns to ignore.
    """

    class Type(Enum):
        File = "file"
        Directory = "directory"
        String = "string"
        _UrlQuery = "url_query"  # Private type, users should not need this

    name: str
    value: Union[str, ArtifactQuery, URLQuery]
    env_var: Optional[str] = None
    type: Optional[Type] = None
    download: bool = False
    mount: Optional[str] = None
    ignore_patterns: list[str] = field(default_factory=list)

    def __post_init__(self):
        if self.env_var is not None and ENV_NAME_RE.match(self.env_var) is None:
            msg = f"env_var ({self.env_var}) is not a valid environment name for shells"
            raise ValueError(msg)


@dataclass
class MaterializedInput:
    value: str
    type: Optional[Input.Type] = None


@dataclass
class AppSerializationSettings:
    """Runtime settings for creating an AppIDL"""

    org: str
    project: str
    domain: str
    is_serverless: bool

    desired_state: Spec.DesiredState
    materialized_inputs: dict[str, MaterializedInput] = field(default_factory=dict)
    additional_distribution: Optional[str] = None


@dataclass
class InputBackend:
    """
    Input information for the backend.
    """

    name: str
    value: str
    download: bool
    type: Optional[Input.Type] = None
    env_var: Optional[str] = None
    dest: Optional[str] = None
    ignore_patterns: str = field(default_factory=list)

    def __post_init__(self):
        if self.type is None:
            if any(self.value.startswith(proto) for proto in SUPPORTED_FS_PROTOCOLS):
                if _has_file_extension(self.value):
                    self.type = Input.Type.File
                else:
                    self.type = Input.Type.Directory
            else:
                self.type = Input.Type.String

        if self.type == Input.Type.String:
            # Nothing to download for a string
            self.download = False

        # If the type is a file or directory and there is a destination, then we
        # automatically assume it is going to be downloaded
        # TODO: In the future, we may mount this, so there is no need to download it
        # with the runtime.
        if self.type in (Input.Type.File, Input.Type.Directory) and self.dest is not None:
            self.download = True

    @classmethod
    def from_input(cls, user_input: Input, settings: AppSerializationSettings) -> "InputBackend":
        if isinstance(user_input.value, str):
            value = user_input.value
            input_type = user_input.type
        else:
            # ArtifactQuery or URLQuery
            try:
                materialized_input = settings.materialized_inputs[user_input.name]
                value = materialized_input.value
                input_type = materialized_input.type or user_input.type
            except KeyError:
                msg = f"Did not materialize {user_input.name}"
                raise ValueError(msg)

        return InputBackend(
            name=user_input.name,
            value=value,
            download=user_input.download,
            env_var=user_input.env_var,
            type=input_type,
            dest=user_input.mount,
            ignore_patterns=user_input.ignore_patterns,
        )


@dataclass
class ServeConfig:
    """
    Configuration for serve runtime.

    :param code_uri: Location of user code in an object store (s3://...)
    :param user_inputs: User inputs. Passed in by `app.inputs`
    """

    code_uri: str  # location of user code
    inputs: List[InputBackend]


SERVE_CONFIG_ENCODER = JSONEncoder(ServeConfig)


@dataclass
class ResolvedInclude:
    src: str
    dest: str


class ScalingMetric:
    @dataclass(frozen=True)
    class Concurrency:
        """
        Use this to specify the concurrency metric for autoscaling, i.e. the number of concurrent requests at a replica
         at which to scale up.
        """

        val: int

        def __post_init__(self):
            if self.val < 1:
                raise ValueError("Concurrency must be greater than or equal to 1")

        def _to_union_idl(self) -> ConcurrencyIDL:
            return ConcurrencyIDL(target_value=self.val)

    @dataclass
    class RequestRate:
        """
        Use this to specify the request rate metric for autoscaling, i.e. the number of requests per second at a replica
         at which to scale up.
        """

        val: int

        def __post_init__(self):
            if self.val < 1:
                raise ValueError("Request rate must be greater than or equal to 1")

        def _to_union_idl(self) -> RequestRateIDL:
            return RequestRateIDL(target_value=self.val)

    @staticmethod
    def _to_union_idl(metric: Optional[Union[RequestRate, Concurrency]]) -> Optional[ScalingMetricIDL]:
        if metric is None:
            return None
        request_rate = None
        concurrency = None
        if isinstance(metric, ScalingMetric.RequestRate):
            request_rate = metric._to_union_idl()
        if isinstance(metric, ScalingMetric.Concurrency):
            concurrency = metric._to_union_idl()
        return ScalingMetricIDL(
            request_rate=request_rate,
            concurrency=concurrency,
        )

    @staticmethod
    def _validate(metric: Optional[Union[RequestRate, Concurrency]]):
        if metric is None:
            return
        if not isinstance(metric, (ScalingMetric.RequestRate, ScalingMetric.Concurrency)):
            raise ValueError("scaling_metric must be an instance of Concurrency or RequestRate")


@dataclass
class App:
    """
    App specification.

    :param name: The name of the application.
    :param container_image: The container image to use for the application.
    :param args: Entrypoint to start application.
    :param command: Command to start application.
    :param port: Port application listens to. Currently, this must be 8080 and the application
        must listen on 8080.
    :param requests: Compute resource requests for application.
    :param secrets: Secrets that are requested for application.
    :param limits: Compute resource limits for application.
    :param include: Files to include for your application.
    :param inputs: Inputs for the application.
    :param env: Environment variables to set for the application.
    :param min_replicas: Minimum number of replicas (ignore if autoscaling is set).
    :param max_replicas: Maximum number of replicas (ignore if autoscaling is set).
    :param scaledown_after: Time to wait before scaling down a replica after it has been idle.
    :param scaling_metric: Autoscale based on a parameter, e.g. request rate or concurrency
      (others may be added in the future).
    :param cluster_pool: The target cluster_pool where the app should be deployed.
    :param allow_anonymous: Public URL does not require any authentication
    """

    name: str
    container_image: Union[str, ImageSpec, PodTemplate]
    port: Union[int, List[int]]
    limits: Resources
    requests: Resources = field(default_factory=Resources)
    secrets: List[Secret] = field(default_factory=list)
    args: Optional[Union[List[str], str]] = None
    command: Optional[Union[List[str], str]] = None
    min_replicas: int = 0
    max_replicas: int = 1
    scaledown_after: Optional[Union[int, timedelta]] = None
    scaling_metric: Optional[Union[ScalingMetric.Concurrency, ScalingMetric.RequestRate]] = None
    include: List[str] = field(default_factory=list)
    inputs: List[Input] = field(default_factory=list)
    env: dict = field(default_factory=dict)
    cluster_pool: str = "default"
    accelerator: Optional[BaseAccelerator] = None
    allow_anonymous: bool = False

    _include_resolved: Optional[List[ResolvedInclude]] = None

    def _validate_autoscaling_config(self):
        if self.min_replicas < 0:
            raise ValueError("min_replicas must be greater than or equal to 0")
        if self.max_replicas < 1 or self.max_replicas < self.min_replicas:
            raise ValueError(f"max_replicas must be greater than or equal to 1 or min_replicas:{self.min_replicas}")
        if self.scaledown_after is not None:
            if isinstance(self.scaledown_after, int):
                self.scaledown_after = timedelta(seconds=self.scaledown_after)
            if self.scaledown_after.total_seconds() < 0:
                raise ValueError("scaledown_after must be greater than or equal to 0")
        ScalingMetric._validate(self.scaling_metric)

    def _validate_ports(self):
        if isinstance(self.port, int):
            ports = [self.port]
        else:
            ports = self.port

        for port in ports:
            if port in INVALID_APP_PORTS:
                invalid_ports = ", ".join(str(p) for p in INVALID_APP_PORTS)
                msg = f"port {port} is not allowed. Please do not use ports: {invalid_ports}"
                raise ValueError(msg)

    def __post_init__(self):
        self._validate_autoscaling_config()
        self._validate_ports()

        if self.inputs and self.command is not None:
            msg = "To use inputs, set command=None to use the union-serve entrypoint"
            raise ValueError(msg)

        if self.include and self.command is not None:
            msg = "To use include, set command=None to use th union-serve entrypoint"
            raise ValueError(msg)

        if self.args is None and self.command is None:
            msg = "Provide args to start your app"
            raise ValueError(msg)

        match_re = APP_NAME_RE.fullmatch(self.name)
        if not match_re:
            raise ValueError("App name must consist of lower case alphanumeric characters, '-' or '.'")

        APP_REGISTRY.add_app(self)

    @property
    def include_resolved(self) -> List[ResolvedInclude]:
        if not self.include:
            return []

        # Already resolved
        if self._include_resolved is not None:
            return self._include_resolved

        # Not resolved yet, so we resolve based on the cwd. This only happens when using
        # AppRemote programmatically and not from `union deploy`
        self._resolve_include(os.getcwd(), os.getcwd())
        return self._include_resolved

    def _resolve_include(self, app_directory: str, cwd: str) -> "App":
        """Resolve include based on the working_dir.

        If a path in `include` is prefixed with "./", then those files are
        assumed to be relative to the file that has the App object.
        """
        relative_prefix = "./"
        seen_dests = set()

        included_resolved = []

        for file in self.include:
            normed_file = os.path.normpath(file)
            if file.startswith(relative_prefix):
                # File is relative to the app_directory:
                src_dir = app_directory
            else:
                src_dir = cwd

            src = os.path.join(src_dir, normed_file)

            if "*" in src:
                new_srcs = glob.glob(src, recursive=True)
            else:
                new_srcs = [src]

            for new_src in new_srcs:
                dest = os.path.relpath(new_src, src_dir)
                if platform.system() == "Windows":
                    # The runtime is linux so the destinations need to be understood by linux.
                    dest = dest.replace(os.sep, "/")

                if dest in seen_dests:
                    msg = f"{dest} is in include multiple times. Please remove one of them."
                    raise ValueError(msg)

                seen_dests.add(dest)

                included_resolved.append(ResolvedInclude(src=new_src, dest=dest))

        self._include_resolved = included_resolved
        return self

    def _get_image(self, container_image: Union[str, ImageSpec], settings: AppSerializationSettings) -> str:
        if isinstance(container_image, str):
            if settings.is_serverless and not is_union_image(container_image):
                # Serverless expects the image to be built by the serverless runtime.
                # If the image isn't a union image, then we need to build it.
                container_image = ImageSpec(name=get_image_name(container_image), base_image=container_image)
            else:
                return container_image

        image_spec = container_image
        # Build for image Spec
        from flytekit.image_spec.image_spec import ImageBuildEngine

        if settings.is_serverless and image_spec.packages is not None:
            image_spec = deepcopy(image_spec)
            new_packages = _convert_union_runtime_to_serverless(image_spec.packages)
            image_spec.packages = new_packages

        # TODO: Remove this patching
        # Do not install flytekit or union when building image
        with patch_get_flytekit_for_pypi(), patch_get_unionai_for_pypi():
            ImageBuildEngine.build(image_spec)
        return image_spec.image_name()

    def _get_args(self) -> List[str]:
        if self.args is None:
            return []
        elif isinstance(self.args, str):
            return shlex.split(self.args)
        else:
            # args is a list
            return self.args

    def _get_command(self, settings: AppSerializationSettings) -> List[str]:
        if self.command is None:
            cmd = ["union-serve"]

            serve_config = ServeConfig(
                code_uri=settings.additional_distribution,
                inputs=[InputBackend.from_input(user_input, settings) for user_input in self.inputs],
            )

            cmd.extend(["--config", SERVE_CONFIG_ENCODER.encode(serve_config)])

            return cmd + ["--"]
        elif isinstance(self.command, str):
            return shlex.split(self.command)
        else:
            # args is a list
            return self.command

    def _get_resources(self) -> ResourcesIDL:
        requests = []
        limits = []

        if self.requests.cpu is not None:
            requests.append(
                ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.CPU, value=str(self.requests.cpu))
            )
        if self.requests.mem is not None:
            requests.append(
                ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.MEMORY, value=str(self.requests.mem))
            )
        if self.requests.gpu is not None:
            requests.append(
                ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.GPU, value=str(self.requests.gpu))
            )
        if self.requests.ephemeral_storage is not None:
            requests.append(
                ResourcesIDL.ResourceEntry(
                    name=ResourcesIDL.ResourceName.EPHEMERAL_STORAGE, value=str(self.requests.ephemeral_storage)
                )
            )

        if self.limits.cpu is not None:
            limits.append(ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.CPU, value=str(self.limits.cpu)))
        if self.limits.mem is not None:
            limits.append(ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.MEMORY, value=str(self.limits.mem)))
        if self.limits.gpu is not None:
            limits.append(ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.GPU, value=str(self.limits.gpu)))
        if self.limits.ephemeral_storage is not None:
            limits.append(
                ResourcesIDL.ResourceEntry(
                    name=ResourcesIDL.ResourceName.EPHEMERAL_STORAGE, value=str(self.limits.ephemeral_storage)
                )
            )

        return ResourcesIDL(requests=requests, limits=limits)

    def _get_container(self, settings: AppSerializationSettings) -> Container:
        if isinstance(self.port, int):
            ports = [self.port]
        else:
            ports = self.port
        container_ports = [ContainerPort(container_port=p) for p in ports]

        return Container(
            image=self._get_image(self.container_image, settings),
            command=self._get_command(settings),
            args=self._get_args(),
            resources=self._get_resources(),
            ports=container_ports,
            env=[KeyValuePair(key=k, value=v) for k, v in self.env.items()],
        )

    def _get_extended_resources(self) -> Optional[ExtendedResources]:
        if self.accelerator is None:
            return None
        return ExtendedResources(gpu_accelerator=self.accelerator.to_flyte_idl())

    def _to_union_idl(self, settings: AppSerializationSettings) -> AppIDL:
        security_context_kwargs = {}
        security_context = None
        if self.secrets:
            security_context_kwargs["secrets"] = [s.to_flyte_idl() for s in self.secrets]
        if self.allow_anonymous:
            security_context_kwargs["allow_anonymous"] = self.allow_anonymous

        if security_context_kwargs:
            security_context = SecurityContext(**security_context_kwargs)

        scaling_metric = ScalingMetric._to_union_idl(self.scaling_metric)

        dur = None
        if self.scaledown_after:
            from google.protobuf.duration_pb2 import Duration

            dur = Duration()
            dur.FromTimedelta(self.scaledown_after)

        autoscaling = AutoscalingConfig(
            replicas=Replicas(min=self.min_replicas, max=self.max_replicas),
            scaledown_period=dur,
            scaling_metric=scaling_metric,
        )

        spec_kwargs = {}
        if isinstance(self.container_image, (str, ImageSpec)):
            spec_kwargs["container"] = self._get_container(settings)
        elif isinstance(self.container_image, PodTemplate):
            spec_kwargs["pod"] = self._get_k8s_pod(self.container_image, settings)
        else:
            msg = "container_image must be a str, ImageSpec or PodTemplate"
            raise ValueError(msg)

        return AppIDL(
            metadata=Meta(
                id=Identifier(
                    org=settings.org,
                    project=settings.project,
                    domain=settings.domain,
                    name=self.name,
                ),
            ),
            spec=Spec(
                desired_state=settings.desired_state,
                ingress=IngressConfig(private=False),
                autoscaling=autoscaling,
                security_context=security_context,
                cluster_pool=self.cluster_pool,
                extended_resources=self._get_extended_resources(),
                **spec_kwargs,
            ),
        )

    def _get_k8s_pod(self, pod_template: PodTemplate, settings: AppSerializationSettings) -> K8sPod:
        """Convert pod_template into a K8sPod IDL."""
        import json

        from google.protobuf.json_format import Parse
        from google.protobuf.struct_pb2 import Struct

        pod_spec_dict = self._serialized_pod_spec(pod_template, settings)
        pod_spec_idl = Parse(json.dumps(pod_spec_dict), Struct())

        metadata = K8sObjectMetadata(
            labels=pod_template.labels,
            annotations=pod_template.annotations,
        )
        return K8sPod(pod_spec=pod_spec_idl, metadata=metadata)

    @staticmethod
    def _sanitize_resource_name(resource: ResourcesIDL.ResourceEntry) -> str:
        return ResourcesIDL.ResourceName.Name(resource.name).lower().replace("_", "-")

    def _serialized_pod_spec(
        self,
        pod_template: PodTemplate,
        settings: AppSerializationSettings,
    ) -> dict:
        """Convert pod spec into a dict."""
        from kubernetes.client import ApiClient
        from kubernetes.client.models import V1Container, V1ContainerPort, V1EnvVar, V1ResourceRequirements

        pod_template = copy.deepcopy(pod_template)

        if pod_template.pod_spec is None:
            return {}

        if pod_template.primary_container_name != "app":
            msg = "Primary container name must be 'app'"
            raise ValueError(msg)

        containers: list[V1Container] = pod_template.pod_spec.containers
        primary_exists = any(container.name == pod_template.primary_container_name for container in containers)

        if not primary_exists:
            msg = "Primary container does not exist with name 'app' does not exist"
            raise ValueError(msg)

        final_containers = []

        # Move all resource names into containers
        for container in containers:
            container.image = self._get_image(container.image, settings)

            if container.name == pod_template.primary_container_name:
                container.args = self._get_args()
                container.command = self._get_command(settings)

                limits, requests = {}, {}
                resources = self._get_resources()
                for resource in resources.limits:
                    limits[self._sanitize_resource_name(resource)] = resource.value
                for resource in resources.requests:
                    requests[self._sanitize_resource_name(resource)] = resource.value

                resource_requirements = V1ResourceRequirements(limits=limits, requests=requests)

                if limits or requests:
                    container.resources = resource_requirements

                if self.env:
                    container.env = [V1EnvVar(name=k, value=v) for k, v in self.env.items()] + (container.env or [])

                if isinstance(self.port, int):
                    ports = [self.port]
                else:
                    ports = self.port
                container.ports = [V1ContainerPort(container_port=p) for p in ports] + (container.ports or [])

            final_containers.append(container)

        pod_template.pod_spec.containers = final_containers
        return ApiClient().sanitize_for_serialization(pod_template.pod_spec)

    def query_endpoint(self, *, public: bool) -> URLQuery:
        """
        Query for endpoint.

        :param public: Whether to return the public or internal endpoint.
        :returns: Object representing a URL query.
        """
        return URLQuery(name=self.name, public=public)
